use crate::schema::{images, tags};
use anyhow::{Context, Result};
use chrono::NaiveDateTime;
use diesel::{
    deserialize::FromSqlRow,
    expression::{helper_types::AsExprOf, AsExpression},
    row::Row,
    sql_types::{Integer, Text},
    sqlite::Sqlite,
    Queryable,
};
use std::convert::TryFrom;
use std::error::Error;
use std::{path::Path, time::UNIX_EPOCH};
use strum_macros::{Display, EnumString};
use uuid::Uuid;

#[derive(Debug, Clone, Queryable, Identifiable, Insertable)]
pub struct Image {
    pub id: String,
    pub path: String,
    pub title: Option<String>,
    pub image_type: ImageType,
    pub last_changed: NaiveDateTime,
    pub published: bool,
}

impl Image {
    pub fn create_uuid(path: &Path) -> anyhow::Result<Uuid> {
        let abs_path = path.canonicalize().context("Path not valid!")?;
        let os_path = abs_path.clone().into_os_string();
        let path_str = os_path.into_string().expect(&format!(
            "Converting path-name problem! {:?}",
            path.display()
        ));
        Ok(Uuid::new_v5(&Uuid::NAMESPACE_URL, &path_str.as_bytes()))
    }

    pub fn try_from_path(path: &Path) -> anyhow::Result<Self> {
        let abs_path = path.canonicalize().context("Path not valid!")?;
        let os_path = abs_path.clone().into_os_string();
        let path_str = os_path.into_string().expect(&format!(
            "Converting path-name problem! {:?}",
            path.display()
        ));
        let uuid = Uuid::new_v5(&Uuid::NAMESPACE_URL, &path_str.as_bytes());
        let metadata = abs_path
            .metadata()
            .context(format!("Unable to load Metadata for: {}", path.display()))?;
        let time = match metadata.modified() {
            Ok(t) => t,
            Err(_) => metadata.created().context(format!(
                "Unable to read created-timestamp of: {}",
                path.display()
            ))?,
        };
        let mtime = time.duration_since(UNIX_EPOCH).unwrap();
        let image_type = ImageType::from_path(path);
        Ok(Image {
            id: uuid.to_string(),
            path: path_str.to_string(),
            title: None,
            image_type,
            last_changed: NaiveDateTime::from_timestamp(
                mtime.as_secs() as i64,
                mtime.subsec_nanos() as u32,
            ),
            published: false,
        })
    }
}

#[derive(Debug, Clone, Queryable, Identifiable, Insertable)]
#[primary_key(image_id, name)]
pub struct Tag {
    pub image_id: String,
    pub tag_type: TagType,
    pub name: String,
}

impl Tag {
    pub fn create_tags(rel_path: &Path) -> Vec<Self> {
        match Image::create_uuid(rel_path) {
            Ok(uuid) => {
                let image_id = uuid.to_string();
                rel_path
                    .ancestors()
                    .filter(|&p| p.display().to_string().len() > 0)
                    .map(|p: &Path| Self {
                        image_id: image_id.clone(),
                        tag_type: TagType::Path,
                        name: p.display().to_string(),
                    })
                    .collect()
            }
            Err(_) => vec![],
        }
    }
}

impl TryFrom<&str> for TagType {
    type Error = String;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match value {
            "Location" => Ok(TagType::Location),
            "Path" => Ok(TagType::Path),
            "Face" => Ok(TagType::Face),
            x => Err(format!("unknown TagType '{}'", x)),
        }
    }
}

#[derive(Clone, Copy, Debug, Display, Eq, PartialEq, SqlType, AsExpression, FromSqlRow)]
pub enum ImageType {
    UNKNOWN,
    JPEG,
    PNG,
    GIF,
    TIFF,
    RAW,
    HEIF,
}

impl ImageType {
    pub fn from_path(path: &Path) -> Self {
        let ext = path.extension();
        if ext == None {
            return ImageType::UNKNOWN;
        }
        match path.extension().unwrap().to_str() {
            Some("heif") => ImageType::HEIF,
            Some("hevc") => ImageType::HEIF,
            Some("jpeg") => ImageType::JPEG,
            Some("jpg") => ImageType::JPEG,
            Some("png") => ImageType::PNG,
            Some("raw") => ImageType::RAW,
            Some("tif") => ImageType::TIFF,
            Some("tiff") => ImageType::TIFF,
            _ => ImageType::UNKNOWN,
        }
    }

    pub fn to_mimetype(&self) -> &str {
        match self {
            Self::HEIF => "image/heif",
            Self::GIF => "image/gif",
            Self::JPEG => "image/jpeg",
            Self::PNG => "image/png",
            Self::TIFF => "image/tiff",
            _ => "application/octet-stream",
        }
    }
}

impl From<i32> for ImageType {
    fn from(int: i32) -> Self {
        match int {
            1 => ImageType::JPEG,
            2 => ImageType::PNG,
            3 => ImageType::GIF,
            4 => ImageType::TIFF,
            5 => ImageType::RAW,
            6 => ImageType::HEIF,
            _ => ImageType::UNKNOWN,
        }
    }
}

impl From<&ImageType> for i32 {
    fn from(image_type: &ImageType) -> Self {
        match image_type {
            ImageType::UNKNOWN => 0,
            ImageType::JPEG => 1,
            ImageType::PNG => 2,
            ImageType::GIF => 3,
            ImageType::TIFF => 4,
            ImageType::RAW => 5,
            ImageType::HEIF => 6,
        }
    }
}

impl From<ImageType> for i32 {
    fn from(image_type: ImageType) -> Self {
        i32::from(&image_type)
    }
}

// traits for diesel

impl diesel::Queryable<Integer, Sqlite> for ImageType {
    type Row = i32;
    fn build(row: Self::Row) -> Self {
        row.into()
    }
}

impl FromSqlRow<Integer, Sqlite> for ImageType {
    fn build_from_row<R: Row<Sqlite>>(row: &mut R) -> Result<Self, Box<dyn Error + Send + Sync>> {
        Ok(i32::build_from_row(row)?.into())
    }
}

impl AsExpression<Integer> for ImageType {
    type Expression = AsExprOf<i32, Integer>;
    fn as_expression(self) -> Self::Expression {
        <i32 as AsExpression<Integer>>::as_expression(self as i32)
    }
}

impl<'a> AsExpression<Integer> for &'a ImageType {
    type Expression = AsExprOf<i32, Integer>;
    fn as_expression(self) -> Self::Expression {
        <i32 as AsExpression<Integer>>::as_expression(self.into())
    }
}

#[derive(
    Clone, Copy, Debug, Display, Eq, EnumString, PartialEq, SqlType, AsExpression, FromSqlRow,
)]
#[sqlite_type = "Text"]
pub enum TagType {
    Path,
    Face,
    Location,
}

// traits to help diesel.
impl diesel::Queryable<Text, Sqlite> for TagType {
    type Row = String;
    fn build(row: Self::Row) -> Self {
        TagType::try_from(row.as_str()).unwrap()
    }
}

impl AsExpression<Text> for TagType {
    type Expression = AsExprOf<String, Text>;
    fn as_expression(self) -> Self::Expression {
        <String as AsExpression<Text>>::as_expression(self.to_string())
    }
}

impl<'a> AsExpression<Text> for &'a TagType {
    type Expression = AsExprOf<String, Text>;
    fn as_expression(self) -> Self::Expression {
        <String as AsExpression<Text>>::as_expression(self.to_string())
    }
}
